import logging

from flask import request
from flask_restful import Resource, abort
from werkzeug.exceptions import BadRequest

from zmq_marshall import ZMQCommsManager
from electric.models import Operation
from electric.models import ObjectNotFoundException, SystemStorage, Preset, PresetIndex

logger = logging.getLogger('electric.app.{0}'.format(__name__))
comms = ZMQCommsManager()

RETRY_LIMIT = 1


def connection_state_dict(exc=None):
    """
    Returns a dict that wraps up the information provided by the exception as well as
    the connection state of the charger
    """

    state = "connected"
    if exc is not None and isinstance(exc, Exception):
        state = "disconnected"

    value = {
        "charger_presence": state
    }

    if exc is not None:
        value.update({"exception": str(exc)})

    return value


def exception_wrapper(func):
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)

        except ObjectNotFoundException as e:
            abort(404, message=e.message)

        except BadRequest as badRequest:
            # Just return it, it's a validation failure
            raise badRequest

        except Exception, ex:
            return connection_state_dict(ex), 504

    return wrapper


class StatusResource(Resource):
    @exception_wrapper
    def get(self):
        info = comms.get_device_info()

        obj = info.to_primitive()
        obj.update(connection_state_dict())

        return obj


class DialogCloseResource(Resource):
    @exception_wrapper
    def put(self, channel_id):
        channel = int(channel_id)
        if not (channel == 0 or channel == 1):
            return connection_state_dict("Channel number must be 0 or 1"), 403
        # comms.close_messagebox(channel)

        obj = comms.get_channel_status(channel).to_primitive()
        obj.update(connection_state_dict())

        return obj


class ChannelResource(Resource):
    @exception_wrapper
    def get(self, channel_id):
        try:
            channel = int(channel_id)
            if not (channel == 0 or channel == 1):
                return connection_state_dict("Channel number must be 0 or 1"), 403

            # get device status, so we know more about channel state
            device_info = comms.get_device_info()
            status = comms.get_channel_status(channel, device_info.device_id)
            if status is not None:
                if device_info is not None:
                    if channel == 0:
                        status.status = device_info.ch1_status
                    elif channel == 1:
                        status.status = device_info.ch2_status

                obj = status.to_primitive()
                obj.update(connection_state_dict())

                return obj

            return connection_state_dict("No status object returned from get_channel_status call"), 403
        except Exception, e:
            raise e


class ControlRegisterResource(Resource):
    @exception_wrapper
    def get(self):
        control = comms.get_control_register()

        # note: intentionally no connection state
        return control.to_primitive()


class ChargeResource(ControlRegisterResource):
    @exception_wrapper
    def put(self, channel_id, preset_memory_slot):
        device_status = comms.run_operation(Operation.Charge, int(channel_id), int(preset_memory_slot))
        comms.turn_off_logging()
        annotated_device_status = device_status.to_primitive()
        annotated_device_status.update(connection_state_dict())
        return annotated_device_status


class DischargeResource(ControlRegisterResource):
    @exception_wrapper
    def put(self, channel_id, preset_memory_slot):
        device_status = comms.run_operation(Operation.Discharge, int(channel_id), int(preset_memory_slot))
        annotated_device_status = device_status.to_primitive()
        annotated_device_status.update(connection_state_dict())
        return annotated_device_status


class StoreResource(ControlRegisterResource):
    @exception_wrapper
    def put(self, channel_id, preset_memory_slot):
        device_status = comms.run_operation(Operation.Storage, int(channel_id), int(preset_memory_slot))
        annotated_device_status = device_status.to_primitive()
        annotated_device_status.update(connection_state_dict())
        return annotated_device_status


class BalanceResource(ControlRegisterResource):
    @exception_wrapper
    def put(self, channel_id, preset_memory_slot):
        device_status = comms.run_operation(Operation.Balance, int(channel_id), int(preset_memory_slot))
        annotated_device_status = device_status.to_primitive()
        annotated_device_status.update(connection_state_dict())
        return annotated_device_status


class MeasureIRResource(ControlRegisterResource):
    @exception_wrapper
    def put(self, channel_id):
        device_status = comms.measure_ir(int(channel_id))
        annotated_device_status = device_status.to_primitive()
        annotated_device_status.update(connection_state_dict())
        return annotated_device_status


class StopResource(ControlRegisterResource):
    @exception_wrapper
    def put(self, channel_id):
        channel_number = int(channel_id)
        # We do this twice. Once to stop. 2nd time to get past the "STOPS" screen.
        operation_response = comms.stop_operation(channel_number).to_primitive()
        operation_response = comms.stop_operation(channel_number).to_primitive()
        operation_response.update(connection_state_dict())
        return operation_response


class SystemStorageResource(Resource):
    @exception_wrapper
    def get(self):
        syst = comms.get_system_storage()
        obj = syst.to_primitive()
        obj.update(connection_state_dict())

        return obj

    @exception_wrapper
    def put(self):
        json_dict = request.json
        del json_dict['charger_presence']
        system_storage_object = SystemStorage(json_dict)
        return comms.save_system_storage(system_storage_object)


class PresetResource(Resource):
    @exception_wrapper
    def get(self, preset_memory_slot):
        preset_memory_slot = int(preset_memory_slot)
        preset = comms.get_preset(preset_memory_slot)
        return preset.to_primitive()

    @exception_wrapper
    def delete(self, preset_memory_slot):
        # This will only, I think ... work for "at the end"
        preset_memory_slot = int(preset_memory_slot)
        logger.info("Try to delete preset at memory slot {0}".format(preset_memory_slot))
        return comms.delete_preset_at_index(preset_memory_slot)

    @exception_wrapper
    def put(self, preset_memory_slot):
        preset_memory_slot = int(preset_memory_slot)
        json_dict = request.json

        # Turn it into a Preset object
        preset = Preset(json_dict)

        logger.info("Asked to save preset to mem slot: {0} with {1}".format(preset_memory_slot, json_dict))
        return comms.save_preset_to_memory_slot(preset, preset_memory_slot)


class AddNewPresetResource(Resource):
    @exception_wrapper
    def put(self):
        json_dict = request.json

        # Turn it into a Preset object
        preset = Preset(json_dict)

        logger.info("Asked to add a new preset: {0}".format(json_dict))
        return comms.add_new_preset(preset).to_native()


class PresetListResource(Resource):
    @exception_wrapper
    def get(self):
        preset_list = comms.get_full_preset_list()
        # TODO: Error handling

        all_presets = []
        for index in preset_list.range_of_presets():
            # Preset.index is the memory slot it's in, not the position within the index
            memory_slot_number = preset_list.indexes[index]
            preset = comms.get_preset(memory_slot_number)
            if preset.is_used or preset.is_fixed:
                try:
                    native = preset.to_native()
                    all_presets.append(native)
                except Exception, e:
                    logger.info("BOOM: " + e)

        return all_presets

    @exception_wrapper
    def post(self):
        pass


class PresetOrderResource(Resource):
    @exception_wrapper
    def get(self):
        preset_list = comms.get_full_preset_list()
        return preset_list.to_native()

    @exception_wrapper
    def post(self):
        json_dict = request.json
        preset_list = PresetIndex(json_dict)
        return comms.save_full_preset_list(preset_list)





import zmq, logging, sys, os
import zmq.utils.win32

from comms_layer import ChargerCommsManager
from router import route_message
from electric.models import testing_control

logger = logging.getLogger('electric.worker')

# construct ZMQ context for comms, with a single REQUEST endpoint
ctx = zmq.Context()
socket = ctx.socket(zmq.REP)

listen_on = os.environ.get("ELECTRIC_WORKER_LISTEN", "tcp://0.0.0.0:5001")
socket.bind(listen_on)

poller = zmq.Poller()
poller.register(socket, zmq.POLLIN)

charger = ChargerCommsManager()


def stop_application():
    socket.close()
    ctx.term()
    sys.exit(0)


def run_worker():
    logging.basicConfig(level=logging.WARN, format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')

    with zmq.utils.win32.allow_interrupt(stop_application):
        logger.warn("iCharger USB reader worker listening on: %s", listen_on)

        try:
            while True:
                socks = dict(poller.poll())

                if socket in socks:
                    message = socket.recv_pyobj()

                    if "method" not in message:
                        logger.warn("method name not specified, ignoring {0}".format(message))
                        message["exception"] = IOError("no method specified in message - rejecting request")
                    else:
                        method = message["method"]

                        args = None
                        if "args" in message:
                            args = message["args"]

                        if "testing-control" in message:
                            tc = message["testing-control"]
                            testing_control.usb_device_present = tc.usb_device_present
                            testing_control.modbus_read_should_fail = tc.modbus_read_should_fail
                            testing_control.modbus_write_should_fail = tc.modbus_write_should_fail

                        message_log = "message: {0}/{1} with args: {2}".format(message["tag"], method, args)

                        try:
                            logger.info("executing message: {0}".format(message_log))
                            message["response"] = route_message(charger, method, args)
                        except Exception, e:
                            logger.error("EXCEPTION during routing/execution for : {0}, {1}".format(message_log, e))
                            message["raises"] = e

                        try:
                            # potentially; the receiver goes away - before we are finished sending -
                            logger.info("sending response for: {0}".format(message_log))
                            socket.send_pyobj(message)
                            logger.info("sent response for: {0}".format(message_log))
                        except Exception, e:
                            logger.error("unable to send response for {0}, {1}".format(message_log, e))

        except KeyboardInterrupt:
            logger.info("Ctrl-C interrupted worker - aborting...")


if __name__ == "__main__":
    run_worker()


import zmq, os
import logging
from electric.models import TestingControlException, testing_control

logger = logging.getLogger('electric.app.{0}'.format(__name__))
worker_loc = os.environ.get("ELECTRIC_WORKER_CONNECT", "tcp://127.0.0.1:5001")

RESPONSE_TIMEOUT_MS = 10000
# RESPONSE_TIMEOUT_MS = 1000000

class ZMQCommsManager(object):
    """
    The comms manager is responsible for data translation between the charger (via ZMQ) and anyone else.
    Validation is not performed here - the data going in/out is assumed to be correct already.
    """
    locking = False

    def __init__(self):
        self.ctx = zmq.Context()

        self.charger = None
        self.request_tag = 0

        self.poll = zmq.Poller()
        self.open_charger_socket()

    def __del__(self):
        self.close_charger_socket()

    def close_charger_socket(self):
        logger.info("Disconnecting from worker")
        self.poll.unregister(self.charger)
        self.charger.close()
        self.charger = None

    def open_charger_socket(self):
        assert(self.charger is None)
        self.charger = self.ctx.socket(zmq.REQ)
        self.charger.set(zmq.REQ_RELAXED, 1)
        self.charger.set(zmq.REQ_CORRELATE, 1)

        logger.info("Connecting to worker at: %s", worker_loc)
        self.charger.connect(worker_loc)
        self.poll.register(self.charger, zmq.POLLIN)

    def _send_message_get_response(self, method_name, arg_dict = None):
        if testing_control.modbus_read_should_fail:
            raise TestingControlException()
        if testing_control.modbus_write_should_fail:
            raise TestingControlException()

        request = {
            "testing-control": testing_control,
            "method": method_name,
            "tag": self.request_tag
        }

        self.request_tag += 1

        if arg_dict is not None:
            request["args"] = arg_dict

        self.charger.send_pyobj(request)
        sockets = dict(self.poll.poll(RESPONSE_TIMEOUT_MS))
        if self.charger in sockets:
            e = None
            r = None

            try:
                resp = self.charger.recv_pyobj()

                if "raises" in resp:
                    e = resp["raises"]
                    logger.error("message received from ZMQ, contained exception: {0}".format(e))
                elif "response" in resp:
                    r = resp["response"]
                else:
                    e = IOError("request received - but no response was found inside, terrible!")

            except Exception, f:
                logger.error("exception received when reading message from ZMQ: {0}".format(e))
                e = f

            if e is not None:
                raise e

            return r
        else:
            raise IOError("request {1} failed to receive a response at all from ZMQ within 10 seconds: {0}".format(method_name, request["tag"]))

    def turn_off_logging(self):
        return self._send_message_get_response("turn_off_logging")

    def get_device_info(self):
        """
        Returns the following information from the iCharger, known as the 'device only reads message'
        :return: a DeviceInfo instance
        """
        return self._send_message_get_response("get_device_info")

    def get_channel_status(self, channel, device_id=None):
        """"
        Returns the following information from the iCharger, known as the 'channel input read only' message:
        :return: ChannelStatus instance
        """
        return self._send_message_get_response("get_channel_status", {
            "channel": channel,
            "device_id": device_id
        })

    def get_control_register(self):
        "Returns the current run state of a particular channel"
        return self._send_message_get_response("get_control_register")

    def set_active_channel(self, channel):
        if channel not in (0, 1):
            return None

        return self._send_message_get_response("set_active_channel", {
            "channel": channel
        })

    def set_beep_properties(self, beep_index=0, enabled=True, volume=5):
        return self._send_message_get_response("set_beep_properties", {
            "beep_index": beep_index,
            "enabled": enabled,
            "volume": volume
        })

    def get_system_storage(self):
        return self._send_message_get_response("get_system_storage")

    def save_system_storage(self, system_storage_object):
        self._send_message_get_response("save_system_storage", {
            "storage": system_storage_object
        })

        return True

    def select_memory_program(self, memory_slot, channel_number=0):
        return self._send_message_get_response("select_memory_program", {
            "memory_slot": memory_slot,
            "channel": channel_number
        })

    def save_full_preset_list(self, preset_list):
        return self._send_message_get_response("save_full_preset_list", {
            "preset_list": preset_list
        })

    def find_preset_with_name(self, named):
        for preset in self.get_all_presets():
            preset_name = preset.name
            lowercase_name = str.lower(preset_name)
            lower_case_wanted = str.lower(named)
            if lowercase_name == lower_case_wanted:
                return preset
        return None

    def get_all_presets(self):
        preset_list = self.get_full_preset_list()
        # TODO: Error handling

        all_presets = []
        for index in preset_list.range_of_presets():
            # Preset.index is the memory slot it's in, not the position within the index
            memory_slot_number = preset_list.indexes[index]
            preset = self.get_preset(memory_slot_number)
            if preset.is_used or preset.is_fixed:
                all_presets.append(preset)

        return all_presets

    def get_full_preset_list(self):
        return self._send_message_get_response("get_full_preset_list")

    def get_preset(self, memory_slot_number):
        return self._send_message_get_response("get_preset", {
            "memory_slot": memory_slot_number
        })

    def delete_preset_at_index(self, preset_memory_slot_number):
        return self._send_message_get_response("delete_preset_at_index", {
            "memory_slot": preset_memory_slot_number
        })

    def add_new_preset(self, preset):
        '''
        This ALWAYS saves a NEW preset. The presets memory_slot is ignored, and it's
        inserted at the end of the preset index list.
        '''
        return self._send_message_get_response("add_new_preset", {
            "preset": preset
        })

    def save_preset_to_memory_slot(self, preset, memory_slot, write_to_flash=True, verify_write=True):
        '''
        This saves an existing preset to memory.
        It does NOT allocate new presets, or insert them into a preset index list
        '''
        return self._send_message_get_response("save_preset_to_memory_slot", {
            "preset": preset,
            "memory_slot": memory_slot,
            "write_to_flash": write_to_flash,
            "verify_write": verify_write
        })

    def stop_operation(self, channel_number):
        return self._send_message_get_response("stop_operation", {
            "channel": channel_number
        })

    def run_operation(self, operation, channel_number, preset_memory_slot_index=0):
        return self._send_message_get_response("run_operation", {
            "op": operation,
            "channel": channel_number,
            "memory_slot": preset_memory_slot_index
        })

    def measure_ir(self, channel_number):
        channel_number = min(1, max(0, channel_number))

        return self._send_message_get_response("measure_ir", {
            "channel": channel_number
        })





from electric.app import application
import logging
logger = logging.getLogger('electric.app.{0}'.format(__name__))

# import faulthandler, signal
# faulthandler.register(signal.SIGUSR1)
#
#
# def debug_signal_handler(signal, frame):
#     del signal
#     del frame
#
#     try:
#         import rpdb2
#         print
#         print
#         print "Starting embedded RPDB2 debugger. Password is 'foobar'"
#         print
#         print
#         rpdb2.start_embedded_debugger("foobar", True, True)
#         rpdb2.setbreak(depth=1)
#         return
#     except StandardError:
#         pass
#
#     try:
#         import code
#         code.interact()
#     except StandardError as ex:
#         print "%r, returning to normal program flow" % ex
#
# try:
#     signal.signal(
#             vars(signal).get("SIGBREAK") or vars(signal).get("SIGUSR2"),
#             debug_signal_handler
#             )
# except ValueError:
#     # Typically: ValueError: signal only works in main thread
#     pass


def run_server():
    opts = {
        "use_reloader": False,
        "use_debugger": False
    }

    from zmq_marshall import worker_loc
    logger.info("Will connect to ELECTRIC_WORKER at: {0}".format(worker_loc))

    application.run(debug=True, host='0.0.0.0', port=5000, **opts)

if __name__ == "__main__":
    run_server()



import logging, threading, time

logger = logging.getLogger('electric.worker.router')
lock = threading.Lock()

from electric.models import testing_control, TestingControlException

class CachedValue(object):
    cache_expiry_seconds = 1.0

    def __init__(self, v = None):
        super(CachedValue, self).__init__()
        self.updated = None
        self.attr = v

    def set_stale(self):
        self.updated = None

    @property
    def value(self):
        return self.attr

    @value.setter
    def value(self, new_value):
        self.attr = new_value
        self.updated = time.time()

    @property
    def is_stale(self):
        if self.updated is None:
            return True

        right_now = time.time()
        difference = right_now - self.updated

        return difference >= CachedValue.cache_expiry_seconds


cached_device_info = CachedValue()
cached_channel_status = [CachedValue(), CachedValue()]


#
# NOTE: this caches ONLY the GET requests, not the other uses of get_device_info() that are internal to the operation
# of the run_op/stop_op methods.
#
def get_device_info_cached(charger):
    if cached_device_info.is_stale:
        value = charger.get_device_info()
        cached_device_info.value = value
    else:
        value = cached_device_info.value
    return value


def get_channel_status_cached(charger, channel, device_id):
    channel = int(channel)
    assert (channel == 0 or channel == 1)

    if cached_channel_status[channel].is_stale:
        value = charger.get_channel_status(channel, device_id)
        cached_channel_status[channel].value = value
    else:
        value = cached_channel_status[channel].value
    return value


def route_message(charger, method, args):
    if not lock.acquire(False):
        raise threading.ThreadError("request being processed by worker - but there is already one in-flight - worldly assumptions have been damaged")



    try:
        if method == "get_device_info":
            return get_device_info_cached(charger)
        elif method == "get_channel_status":
            return get_channel_status_cached(charger, args["channel"], args["device_id"])
        elif method == "get_control_register":
            return charger.get_control_register()
        elif method == "set_active_channel":
            return charger.set_active_channel(args["channel"])
        elif method == "get_system_storage":
            return charger.get_system_storage()
        elif method == "save_system_storage":
            return charger.save_system_storage(args["storage"])
        elif method == "select_memory_program":
            return charger.select_memory_program(args["memory_slot"], args["channel"])
        elif method == "save_full_preset_list":
            return charger.save_full_preset_list(args["preset_list"])
        elif method == "get_full_preset_list":
            return charger.get_full_preset_list()
        elif method == "get_preset":
            return charger.get_preset(args["memory_slot"])
        elif method == "delete_preset_at_index":
            return charger.delete_preset_at_index(args["memory_slot"])
        elif method == "add_new_preset":
            return charger.add_new_preset(args["preset"])
        elif method == "save_preset_to_memory_slot":
            return charger.save_preset_to_memory_slot(args["preset"], args["memory_slot"], args["write_to_flash"], args["verify_write"])
        elif method == "stop_operation":
            return charger.stop_operation(args["channel"])
        elif method == "run_operation":
            return charger.run_operation(args["op"], args["channel"], args["memory_slot"])
        elif method == "measure_ir":
            return charger.measure_ir(args["channel"])
        elif method == "turn_off_logging":
            return charger.turn_off_logging()
        elif method == "set_beep_properties":
            return charger.set_beep_properties(args["beep_index"], args["enabled"], args["volume"])
        else:
            raise IOError("Unknown method name, cannot execute anything: {0}".format(method))

    finally:
        lock.release()


import copy
import logging
import struct

from schematics.models import Model
from schematics.transforms import blacklist
from schematics.types import StringType, IntType, LongType, FloatType, BooleanType
from schematics.types.compound import ModelType, ListType
from schematics.types.serializable import serializable

logger = logging.getLogger('electric.app.{0}'.format(__name__))


class TestingControl:
    def __init__(self):
        self.reset()

    def reset(self):
        # Holds the global testing flags that modify the interface behaviour in simple ways to enable testing
        self.usb_device_present = True
        # If a read operation should fail and throw an exception
        self.modbus_read_should_fail = False
        # If a write operation should fail and throw an exception
        self.modbus_write_should_fail = False


class TestingControlException(Exception):
    pass

testing_control = TestingControl()

# copied from modbus protocol spec
READ_HOLDING_REGISTERS = 3
READ_INPUT_REGISTERS = 4
WRITE_MULTIPLE_REGISTERS = 16


class Operation:
    Charge = 0
    Storage = 1
    Discharge = 2
    Cycle = 3
    Balance = 4
    MeasureIR = 5


class Order:
    Stop = 0
    Run = 1
    Modify = 2
    WriteSys = 3
    WriteMemHead = 4
    WriteMem = 5
    TransLogOn = 6
    TransLogOff = 7
    MsgBoxYes = 8
    MsgBoxNo = 9


class ChemistryType:
    LiPo = 0
    LiLo = 1
    LiFe = 2
    NiMH = 3
    NiCd = 4
    Pb = 5
    NiZn = 6


'''

DeviceInfo:
- General Status Word:

Channel:
- control_status
- run_status

'''

# "general status word"
# fully stopped / idle: cell volt (no run or control flags. run_status=0 and control_status=0)
# charging: cell volt, control, run
# stopped (after some operation): cell volt, run
# storage (charging): cell volt, control, run
# discharging: ctrl_status, run, run_status

STATUS_RUN = 0x01
STATUS_ERROR = 0x02
STATUS_CONTROL_STATUS = 0x04
STATUS_RUN_STATUS = 0x08
STATUS_DLG_BOX_STATUS = 0x10
STATUS_CELL_VOLTAGE = 0x20
STATUS_BALANCE = 0x40

DEVICEID_4010_DUO = 64
DEVICEID_308_DUO = 66
DEVICEID_306_DUO = 67  # TODO: Update to the real device number


class ObjectNotFoundException(Exception):
    status_code = 404


class BadRequestException(Exception):
    status_code = 404


DeviceIdCellCount = (
    (DEVICEID_308_DUO, 8),
    (DEVICEID_306_DUO, 6),
    (DEVICEID_4010_DUO, 10)
)

Control_RunOperations = (
    (0, "charge"),
    (1, "storage"),
    (2, "discharge"),
    (3, "cycle"),
    (4, "balance only"),
)

Control_OrderOperations = (
    (0, "run"),
    (1, "modify"),
    (2, "write system"),
    (3, "write memory head"),
    (4, "write memory"),
    (5, "trans log on"),
    (6, "trans log off"),
    (7, "msgbox yes"),
    (8, "msgbox no")
)


class ReadDataSegment:
    """
    The iCharger USB HID API cannot read more than 64 bytes at a time, yet some of the data structures are much
    larger than this.  To accommodate multiple reads this class captures all the information about a segment of
    the data-structure, from the base address from which to read, the packed-byte structure (aka C struct) and
    auto calculates the correct offset based off a prior read if required.
    """

    def __init__(self, charger, name, fmt, base=None, prev_format=None):
        self.func_code = READ_INPUT_REGISTERS

        self.name = name
        self.format = fmt
        self.size = struct.calcsize("=" + self.format)
        self.addr = base if base is not None else prev_format.addr + prev_format.size / 2

        if self.addr >= 0x8000:
            self.func_code = READ_HOLDING_REGISTERS

        self.data = charger.modbus_read_registers(self.addr, self.format, function_code=self.func_code)


class WriteDataSegment(object):
    def __init__(self, charger, name, data_tuples, data_format, base=None, prev_format=None):
        self.func_code = WRITE_MULTIPLE_REGISTERS
        self.name = name

        self.data = self.convert_tuples_to_u16s(data_tuples, data_format)

        # / 2 because offsets are expressed in words
        self.addr = base if base is not None else prev_format.addr + (prev_format.size_in_bytes / 2)
        (number_of_words_returned,) = charger.modbus_write_registers(self.addr, self.data)

        # Actually, we know that the data being written is all "H" or "h" (2 bytes)
        # So bytes is simply: len(self.data) * 2
        self.size_in_bytes = len(self.data) * 2

    @staticmethod
    def convert_tuples_to_u16s(data_tuples, data_format):
        # modbus_write_registers expects a bunch of U16s
        # Thus we expect the data_tuples to have a format supplied.
        # We convert this to a set of U16s. As such, each format MUST be of length divisible by 2.
        native_format = "=" + data_format
        format_length = struct.calcsize(native_format)
        if format_length % 2 == 1:
            msg = "Data format {1} has size {0}. Must be divisible by 2 for U16s conversion to succeed".format(format_length, data_format)
            raise Exception(msg)

        packed_data_as_bytes = struct.pack(native_format, *data_tuples)

        # Some paranoia
        size_of_packed_data = len(packed_data_as_bytes)
        if size_of_packed_data % 2 == 1:
            raise Exception(
                "Packed data has size {0} (which is unexpected, since the previous packing size test passed). Must be divisible by 2 for U16s conversion to succeed".format(
                    size_of_packed_data))

        u16s_repeat = size_of_packed_data / 2
        u16s_format = "{0}H".format(u16s_repeat)
        return struct.unpack(u16s_format, packed_data_as_bytes)


class DeviceInfoStatus(Model):
    value = IntType(required=True, min_value=0, max_value=0x7f)

    class Options:
        # this means that value won't appear in the JSON output by default
        roles = {'default': blacklist('value')}

    def __init__(self, value=0):
        super(DeviceInfoStatus, self).__init__()
        self.value = value

    @serializable
    def run(self):
        return (self.value & STATUS_RUN) == STATUS_RUN

    @serializable
    def err(self):
        return (self.value & STATUS_ERROR) == STATUS_ERROR

    @serializable
    def ctrl_status(self):
        return (self.value & STATUS_CONTROL_STATUS) == STATUS_CONTROL_STATUS

    @serializable
    def run_status(self):
        return (self.value & STATUS_RUN_STATUS) == STATUS_RUN_STATUS

    @serializable
    def dlg_box_status(self):
        return (self.value & STATUS_DLG_BOX_STATUS) == STATUS_DLG_BOX_STATUS

    @serializable
    def cell_volt_status(self):
        return (self.value & STATUS_CELL_VOLTAGE) == STATUS_CELL_VOLTAGE

    @serializable
    def balance(self):
        return (self.value & STATUS_BALANCE) == STATUS_BALANCE


class DeviceInfo(Model):
    device_id = IntType(required=True)
    device_sn = StringType(required=True)
    software_ver = LongType(required=True)
    hardware_ver = LongType(required=True)
    system_len = IntType(required=True)
    memory_len = IntType(required=True)
    channel_count = IntType(required=True, default=2)
    cell_count = IntType(required=True, default=0)
    ch1_status = ModelType(DeviceInfoStatus, default=DeviceInfoStatus())
    ch2_status = ModelType(DeviceInfoStatus, default=DeviceInfoStatus())

    def __init__(self, modbus_data=None):
        super(DeviceInfo, self).__init__()
        if modbus_data is not None:
            self.set_from_modbus_data(modbus_data)

    def get_status(self, channel):
        return self.ch1_status if channel == 0 else self.ch2_status

    def set_from_modbus_data(self, data):
        self.device_id = data[0]
        self.device_sn = data[1].split('\0')[0]
        self.software_ver = data[2]
        self.hardware_ver = data[3]
        self.system_len = data[4]
        self.memory_len = data[5]
        self.ch1_status = DeviceInfoStatus(data[6])
        self.ch2_status = DeviceInfoStatus(data[7])

        for (device_id, cell_count) in DeviceIdCellCount:
            if device_id == self.device_id:
                self.cell_count = cell_count


class OperationResponse(Model):
    first_number = IntType(required=True)
    second_number = IntType(required=False)

    def __init__(self, modbus_data=None):
        super(OperationResponse, self).__init__()
        if modbus_data is not None:
            self.set_from_modbus_data(modbus_data)

    # Beats me
    @serializable()
    def success(self):
        return self.first_number == 128

    @serializable()
    def error(self):
        return not self.success

    def set_from_modbus_data(self, data):
        self.first_number = data[0]
        if len(data) > 1:
            self.second_number = data[1]


class CellStatus(Model):
    cell = IntType(required=True, min_value=0, max_value=9)
    voltage = FloatType(required=True, serialized_name="v")
    balance = IntType(required=True)
    ir = FloatType(required=True)

    def __init__(self, c=0, volt=0, bal=0, i=0):
        super(CellStatus, self).__init__()
        self.set_from_modbus_data(c, volt, bal, i)

    def set_from_modbus_data(self, c, volt, bal, i):
        self.cell = c
        self.voltage = volt / 1000.0
        self.balance = bal
        self.ir = i / 10.0


class ChannelStatus(Model):
    channel = IntType(required=True, min_value=0, max_value=1)
    timestamp = LongType(required=True, default=0)
    curr_out_power = FloatType(required=True)
    curr_out_amps = FloatType(required=True)
    curr_inp_volts = FloatType(required=True)
    curr_out_volts = FloatType(required=True)
    curr_out_capacity = FloatType(required=True)
    curr_int_temp = FloatType(required=True)
    curr_ext_temp = FloatType(required=True)

    cells = ListType(ModelType(CellStatus))

    cell_total_ir = FloatType(required=True)
    cell_total_voltage = FloatType(required=True)
    cell_count_with_voltage_values = FloatType(required=True)
    cycle_count = IntType(required=True)
    control_status = IntType(required=True)
    run_status = IntType(required=True)
    run_error = IntType(required=True)
    dlg_box_id = IntType(required=True)
    line_intern_resistance = FloatType(required=True)

    # Optionally added to the response, when a channel status is requested
    status = ModelType(DeviceInfoStatus)

    @staticmethod
    def modbus(device_id=None, channel=0, header=None, cell_v=None, cell_b=None, cell_i=None, footer=None):
        status = ChannelStatus()
        if header is not None and cell_v is not None and cell_b is not None and cell_i is not None and footer is not None:
            status.set_from_modbus_data(device_id, channel, header, cell_v, cell_b, cell_i, footer)
        return status

    def set_from_modbus_data(self, device_id, channel, data, cell_v, cell_b, cell_i, footer):
        self.channel = channel

        self.timestamp = data[0] / 1000.0
        self.curr_out_power = data[1] / 1000.0
        self.curr_out_amps = data[2] / 100.0
        self.curr_inp_volts = data[3] / 1000.0
        self.curr_out_volts = data[4] / 1000.0
        self.curr_out_capacity = data[5]  # mAh sent or taken from batt
        self.curr_int_temp = data[6] / 10.0
        self.curr_ext_temp = data[7] / 10.0

        self.cell_count_with_voltage_values = 0
        self.cell_total_voltage = 0

        cells = []
        for x in range(0, 10):
            if cell_v[x] == 1024:
                continue
            c = CellStatus(x, cell_v[x], cell_b[x], cell_i[x])
            cells.append(c)
            self.cell_total_voltage += c.voltage
            if c.voltage > 0:
                self.cell_count_with_voltage_values += 1

        self.cells = cells

        self.cell_total_ir = footer[0] / 10.0
        self.line_intern_resistance = footer[1] / 10.0
        self.cycle_count = footer[2]
        self.control_status = footer[3]
        self.run_status = footer[4]
        self.run_error = footer[5]
        self.dlg_box_id = footer[6]

        if device_id:
            # With this, we can work out if the main battery lead is plugged in
            if device_id is DEVICEID_306_DUO:
                max_voltage = 26
            elif device_id is DEVICEID_308_DUO:
                max_voltage = 30
            else:
                max_voltage = 40
            if self.curr_out_volts > max_voltage:
                self.curr_out_volts = 0

        # If the charger isn't doing anything, make sure timestamp shows 00:00
        if self.run_status == 0:
            self.timestamp = 0

    def max_charger_input_voltage(self):
        return 0

    @serializable
    def battery_plugged_in(self):
        total_voltage = self.cell_total_voltage
        curr_out_volts = self.curr_out_volts

        # This is more or less a "punt".
        # You can't use exact, because sometimes total will be > sometimes a little less. Go figure.
        is_plugged_in_threshold = (total_voltage * 0.8)
        plugged_in = curr_out_volts > is_plugged_in_threshold
        if is_plugged_in_threshold > 0 and plugged_in > 0:
            if not plugged_in:
                logger.info("Pack not plugged in because {0}:{1} > {2}:{3}".format(curr_out_volts, type(curr_out_volts), is_plugged_in_threshold, type(is_plugged_in_threshold)))
        return plugged_in

    @serializable
    def balance_leads_plugged_in(self):
        return self.cell_total_voltage > 0


class Control(Model):
    op = IntType(required=True)
    memory = IntType(required=True)
    channel = IntType(required=True, min_value=0, max_value=1)
    order_lock = StringType(required=True)
    order = IntType(default=0)
    limit_current = FloatType(required=True)
    limit_volt = FloatType(required=True)

    def __init__(self, modbus_data=None):
        super(Control, self).__init__()
        if modbus_data is not None:
            self.set_from_modbus_data(modbus_data)

    @serializable
    def op_description(self):
        for (num, name) in Control_RunOperations:
            if self.op == num:
                return name
        return None

    @serializable
    def order_description(self):
        for (num, name) in Control_OrderOperations:
            if self.order == num:
                return name
        return None

    def set_from_modbus_data(self, data):
        self.op = data[0]
        self.memory = data[1]
        self.channel = data[2]
        self.order_lock = "0x%0.4X" % data[3]
        self.order = data[4]
        self.limit_current = data[5] / 1000.0
        self.limit_volt = data[6] / 1000.0


class SystemStorage(Model):
    temp_unit = StringType(required=True, choices=["C", "F"], min_length=1, max_length=1)
    temp_stop = FloatType(required=True, min_value=60, max_value=75)
    temp_fans_on = FloatType(required=True, min_value=30, max_value=50)
    temp_reduce = FloatType(required=True, min_value=5, max_value=20)

    fans_off_delay = IntType(required=True, min_value=0, max_value=10)
    lcd_contrast = IntType(required=True, min_value=0, max_value=32)
    lcd_brightness = IntType(required=True, min_value=1, max_value=33)

    # for the next three the 4010 doesn't provide a way to set these in the UI?
    beep_type_key = IntType(required=True, min_value=0, max_value=0)
    beep_type_hint = IntType(required=True, min_value=0, max_value=0)
    beep_type_alarm = IntType(required=True, min_value=0, max_value=0)
    beep_type_done = IntType(required=True, min_value=0, max_value=3)

    beep_enabled_key = BooleanType(required=True)
    beep_enabled_hint = BooleanType(required=True)
    beep_enabled_alarm = BooleanType(required=True)
    beep_enabled_done = BooleanType(required=True)

    beep_volume_key = IntType(required=True, min_value=0, max_value=10)
    beep_volume_hint = IntType(required=True, min_value=0, max_value=10)
    beep_volume_alarm = IntType(required=True, min_value=0, max_value=10)
    beep_volume_done = IntType(required=True, min_value=0, max_value=10)

    # we shouldn't even expose this?!?
    calibration = IntType(required=True, min_value=0, max_value=1) # factory=0, user=1

    selected_input_source = IntType(required=True)

    dc_input_low_voltage = FloatType(required=True, min_value=9, max_value=48)
    dc_input_over_voltage = FloatType(required=True)
    dc_input_current_limit = FloatType(required=True, min_value=1, max_value=65)

    batt_input_low_voltage = FloatType(required=True, min_value=9, max_value=48)
    batt_input_over_voltage = FloatType(required=True)
    batt_input_current_limit = FloatType(required=True, min_value=1, max_value=65)

    regenerative_enable = IntType(required=True, min_value=0, max_value=1)
    regenerative_volt_limit = FloatType(required=True, min_value=9, max_value=48)
    regenerative_current_limit = FloatType(required=True, min_value=1, max_value=65)

    power_priority = IntType(required=True, min_value=0, max_value=2)

    charge_power = ListType(IntType(min_value=5,  max_value=1400))
    discharge_power = ListType(IntType(min_value=5, max_value=130))
    monitor_log_interval = ListType(IntType)
    monitor_save_to_sd = ListType(BooleanType)

    servo_type = LongType(required=True, min_value=0, max_value=3) # 0:Analog 1500us/50hz, 1:Digital 1500us/333Hz, 2:Digital 760us/560hz, 3:User
    servo_user_center = LongType(required=True, min_value=700, max_value=1600)
    server_user_rate = LongType(required=True, min_value=40, max_value=700)
    server_user_op_angle = LongType(required=True, min_value=100, max_value=1000)

    modbus_mode = LongType(required=True, min_value=0, max_value=1)
    modbus_serial_addr = LongType(required=True, min_value=0, max_value=254)
    modbus_serial_baud_rate = LongType(required=True, min_value=0, max_value=5) # 9600, 19200, 38400, 57600, 115200, 128000
    modbus_serial_parity = LongType(required=True, min_value=0, max_value=2) # None, Even, Odd

    @staticmethod
    def modbus(ds1=None, ds2=None, ds3=None):
        if ds1 is not None and ds2 is not None and ds3 is not None:
            storage = SystemStorage()
            storage.set_from_modbus_data(ds1, ds2, ds3)
            return storage
        return None

    def set_from_modbus_data(self, ds1, ds2, ds3):
        dummy1 = None
        (self.temp_unit, self.temp_stop, self.temp_fans_on, self.temp_reduce, dummy1, self.fans_off_delay,
         self.lcd_contrast, self.lcd_brightness, dummy1,
         self.beep_type_key, self.beep_type_hint, self.beep_type_alarm, self.beep_type_done,
         self.beep_enabled_key, self.beep_enabled_hint, self.beep_enabled_alarm, self.beep_enabled_done,
         self.beep_volume_key, self.beep_volume_hint, self.beep_volume_alarm, self.beep_volume_done) = ds1.data

        (dummy1, self.calibration, dummy1, self.selected_input_source,
         self.dc_input_low_voltage, self.dc_input_over_voltage, self.dc_input_current_limit,
         self.batt_input_low_voltage, self.batt_input_over_voltage, self.batt_input_current_limit,
         self.regenerative_enable, self.regenerative_volt_limit, self.regenerative_current_limit) = ds2.data

        self.charge_power = [0, 0]
        self.discharge_power = [0, 0]
        self.monitor_log_interval = [0, 0]
        self.monitor_save_to_sd = [False, False]

        (self.charge_power[0], self.charge_power[1],
         self.discharge_power[0], self.discharge_power[1],
         self.power_priority,
         self.monitor_log_interval[0], self.monitor_log_interval[1],
         self.monitor_save_to_sd[0], self.monitor_save_to_sd[1],
         self.servo_type, self.servo_user_center, self.server_user_rate, self.server_user_op_angle,
         self.modbus_mode, self.modbus_serial_addr, self.modbus_serial_baud_rate,
         self.modbus_serial_parity) = ds3.data

        self.temp_unit = "C" if self.temp_unit is 0 else "F"

        self.temp_stop /= 10.0
        self.temp_fans_on /= 10.0
        self.temp_reduce /= 10.0

        self.dc_input_low_voltage /= 10.0
        self.dc_input_current_limit /= 10.0
        self.dc_input_over_voltage /= 10.0

        self.batt_input_low_voltage /= 10.0
        self.batt_input_over_voltage /= 10.0
        self.batt_input_current_limit /= 10.0

        self.regenerative_volt_limit /= 10.0
        self.regenerative_current_limit /= 10.0

        self.servo_user_center /= 10
        self.server_user_op_angle /= 10

    def to_modbus_data(self):
        s1 = (0 if self.temp_unit == "C" else 1, self.temp_stop * 10, self.temp_fans_on * 10, self.temp_reduce * 10,)

        # Note, the trailing 0 is required, else the lcd_brightness isn't set correctly.
        s2 = (self.fans_off_delay, self.lcd_contrast, self.lcd_brightness, 0,)
        s3 = (self.beep_type_key, self.beep_type_hint, self.beep_type_alarm, self.beep_type_done,
              self.beep_enabled_key, self.beep_enabled_hint, self.beep_enabled_alarm, self.beep_enabled_done,
              self.beep_volume_key, self.beep_volume_hint, self.beep_volume_alarm, self.beep_volume_done,)
        s4 = (self.calibration,)

        # Note, the trailing 0 is required, else the regenerative_current_limit isn't set correctly.
        s5 = (self.selected_input_source,
              self.dc_input_low_voltage * 10, self.dc_input_over_voltage * 10, self.dc_input_current_limit * 10,
              self.batt_input_low_voltage * 10, self.batt_input_over_voltage * 10, self.batt_input_current_limit * 10,
              self.regenerative_enable, self.regenerative_volt_limit * 10, self.regenerative_current_limit * 10, 0,)

        # Note, the trailing 0 is required, else the modbus_serial_parity isn't set correctly.
        s6 = (self.charge_power[0], self.charge_power[1],
              self.discharge_power[0], self.discharge_power[1],
              self.power_priority,
              self.monitor_log_interval[0], self.monitor_log_interval[1],
              self.monitor_save_to_sd[0], self.monitor_save_to_sd[1],
              self.servo_type, self.servo_user_center * 10, self.server_user_rate, self.server_user_op_angle * 10,
              self.modbus_mode, self.modbus_serial_addr, self.modbus_serial_baud_rate,
              self.modbus_serial_parity, 0)
        return s1, s2, s3, s4, s5, s6

    @serializable
    def selected_input_source_type(self):
        return "dc" if self.selected_input_source == 0 else "battery"

    @serializable
    def power_priority_description(self):
        if self.power_priority == 0:
            return "average"
        if self.power_priority == 1:
            return "ch1 priority"
        if self.power_priority == 2:
            return "ch2 priority"


class PresetIndex(Model):
    indexes = ListType(IntType(min_value=0, max_value=255), required=True, min_size=0, max_size=63, default=[])

    @staticmethod
    def modbus(count=None, indexes=None):
        if count is not None and indexes is not None:
            pi = PresetIndex()
            pi.set_from_modbus_data(count, indexes)
            return pi
        return None

    def preset_exists_at_index(self, index):
        return self.indexes[index] != 255

    def index_of_preset_with_memory_slot_number(self, memory_slot_number):
        for index in self.range_of_presets():
            if self.indexes[index] == memory_slot_number:
                return index
        return None

    @property
    def next_available_memory_slot(self):
        for slot_number in range(0, 64):
            # does any index take this?
            memory_slot_number = self.index_of_preset_with_memory_slot_number(slot_number)
            if memory_slot_number is None:
                return slot_number
        return None

    def add_to_index(self, new_preset):
        next_memory_slot = self.next_available_memory_slot
        if next_memory_slot is None:
            return None

        next_index = self.first_empty_index_position
        if next_index is None:
            return None

        new_preset.memory_slot = next_memory_slot
        new_preset.use_flag = 0x55aa  # Used
        self.indexes[next_index] = next_memory_slot

        return new_preset

    def range_of_presets(self):
        if self.first_empty_index_position is not None:
            return range(0, self.first_empty_index_position)
        return range(0, 64)

    def is_valid_index(self, index):
        return index in self.range_of_presets()

    def swap(self, index_one, index_two):
        if not self.is_valid_index(index_one):
            message = "Index one {0} is invalid".format(index_one)
            raise ObjectNotFoundException(message)
        if not self.is_valid_index(index_two):
            message = "Index two {0} is invalid".format(index_two)
            raise ObjectNotFoundException(message)

        value1 = self.indexes[index_one]
        value2 = self.indexes[index_two]
        self.indexes[index_one] = value2
        self.indexes[index_two] = value1

    @property
    def number_of_presets(self):
        first_empty_slot = self.first_empty_index_position
        if not first_empty_slot:
            return len(self.indexes)
        return first_empty_slot

    @property
    def first_empty_index_position(self):
        first_empty = None
        for i, index in enumerate(self.indexes):
            if index == 255:
                first_empty = i
                break

        return first_empty

    def delete_item_at_index(self, index, validate=True):
        if self.is_valid_index(index):
            del self.indexes[index]
            if validate:
                self._validate_and_fix_index_list()

    def _validate_and_fix_index_list(self):
        # Add 255's on the end until we have 64 of them
        # this also clears out any other 0's and other stuff.
        if self.first_empty_index_position is not None:
            for index in range(self.first_empty_index_position, 64):
                if index < len(self.indexes):
                    self.indexes[index] = 255
                else:
                    self.indexes.append(255)

        # Indexes must be unique
        set_of_seen = {}
        current_position = 0
        while current_position < self.number_of_presets - 1:
            index_at_this_position = self.indexes[current_position]
            if set_of_seen.get(index_at_this_position, None):
                # if we've already seen it, we should delete this one
                self.delete_item_at_index(current_position, validate=False)

                # Don't inc counter, as we've deleted one
            else:
                set_of_seen[index_at_this_position] = 1
                current_position += 1

    def set_indexes(self, new_value):
        self.indexes = copy.deepcopy(new_value)
        self._validate_and_fix_index_list()

    def set_from_modbus_data(self, count, indexes):
        self.set_indexes(indexes)

    @serializable
    def count(self):
        # The number of non-255 index positions
        return self.number_of_presets

    def to_modbus_data(self):
        v1 = [self.number_of_presets, ]
        v1.extend(self.indexes[:32])
        v2 = self.indexes[32:]
        return v1, v2


class Preset(Model):
    # The memory slot that this Preset occupies
    memory_slot = IntType(required=True, min_value=0, max_value=63, serialized_name="index")

    use_flag = LongType(required=True, choices=[0xffff, 0x55aa, 0x0000])
    name = StringType(required=True, max_length=37)
    capacity = LongType(required=True)
    auto_save = BooleanType(required=True, default=False)
    li_balance_end_mode = IntType(required=True, default=0)
    op_enable_mask = IntType(required=True, default=0xff)

    channel_mode = IntType(required=True, choices=[0, 1])
    save_to_sd = BooleanType(required=True, default=True)
    log_interval_sec = FloatType(required=True, min_value=0.5, max_value=60)
    run_counter = IntType(required=True)

    type = IntType(required=True, choices=[0, 1, 2, 3, 4, 5])
    li_cell = IntType(required=True)
    ni_cell = IntType(required=True)
    pb_cell = IntType(required=True, min_value=1, max_value=15)

    li_mode_c = IntType(required=True, choices=[0, 1])
    li_mode_d = IntType(required=True, min_value=1, max_value=3)  # this is a bitmask
    ni_mode_c = IntType(required=True, choices=[0, 1])
    ni_mode_d = IntType(required=True, choices=[0, 1])
    pb_mode_c = IntType(required=True, choices=[0, 1])
    pb_mode_d = IntType(required=True, choices=[0, 1])

    bal_speed = IntType(required=True, choices=[0, 1, 2, 3])  # 0=slow, 1=normal, 2=fast, 3=user
    bal_start_mode = IntType(required=True)
    bal_start_voltage = IntType(required=True)
    bal_diff = IntType(required=True)
    bal_over_point = IntType(required=True)
    bal_set_point = IntType(required=True)
    bal_delay = IntType(required=True)

    keep_charge_enable = BooleanType(required=True)

    lipo_charge_cell_voltage = FloatType(required=True, min_value=3.85, max_value=4.35)
    lilo_charge_cell_voltage = FloatType(required=True, min_value=3.75, max_value=4.35)
    life_charge_cell_voltage = FloatType(required=True, min_value=3.3, max_value=3.8)

    lipo_storage_cell_voltage = FloatType(required=True, min_value=3.7, max_value=3.9)
    lilo_storage_cell_voltage = FloatType(required=True, min_value=3.6, max_value=3.8)
    life_storage_cell_voltage = FloatType(required=True, min_value=3.1, max_value=3.4)

    lipo_discharge_cell_voltage = FloatType(required=True, min_value=3, max_value=4.1)
    lilo_discharge_cell_voltage = FloatType(required=True, min_value=2.5, max_value=4)
    life_discharge_cell_voltage = FloatType(required=True, min_value=2, max_value=3.5)

    charge_current = FloatType(required=True, min_value=0.05, max_value=40)
    discharge_current = FloatType(required=True, min_value=0.05, max_value=40)
    end_charge = FloatType(required=True, min_value=1, max_value=50)
    end_discharge = FloatType(required=True, min_value=1, max_value=100)
    regen_discharge_mode = IntType(required=True, choices=[0, 1, 2])

    ni_peak = FloatType(required=True, min_value=1, max_value=20)
    ni_peak_delay = IntType(required=True, min_value=0, max_value=20)
    ni_trickle_enable = BooleanType(required=True)
    ni_trickle_current = FloatType(required=True, min_value=0.02, max_value=1)
    ni_trickle_time = IntType(required=True, min_value=1, max_value=999)

    ni_zero_enable = BooleanType(required=True)

    ni_discharge_voltage = FloatType(required=True, min_value=0.1, max_value=33)
    pb_charge_voltage = FloatType(required=True, min_value=2, max_value=2.6)
    pb_discharge_voltage = FloatType(required=True, min_value=1.5, max_value=2.4)
    pb_cell_float_enable = BooleanType(required=True)
    pb_cell_float_voltage = FloatType(required=True)

    restore_voltage = FloatType(required=True, min_value=0.5, max_value=2.5)
    restore_time = IntType(required=True, min_value=1, max_value=5)
    restore_current = FloatType(required=True, min_value=0.02, max_value=0.5)

    cycle_count = IntType(required=True, min_value=1, max_value=99)
    cycle_delay = IntType(required=True, min_value=0, max_value=3000)  # charger goes more. I set an arbitrary limit.
    cycle_mode = IntType(required=True, choices=[0, 1, 2, 3, 4, 5])

    safety_time_c = IntType(required=True, min_value=0)
    safety_cap_c = IntType(required=True, min_value=50, max_value=200)
    safety_temp_c = FloatType(required=True, min_value=20, max_value=80)

    safety_time_d = IntType(required=True, min_value=0)
    safety_cap_d = IntType(required=True, min_value=50, max_value=200)
    safety_temp_d = FloatType(required=True, min_value=20, max_value=80)

    reg_ch_mode = IntType(required=True, choices=[0, 1, 2])
    reg_ch_volt = FloatType(required=True, min_value=0.1, max_value=33)
    reg_ch_current = FloatType(required=True, min_value=0.05, max_value=40)

    fast_store = BooleanType(required=True, default=True)
    store_compensation = FloatType(required=True, min_value=0, max_value=0.2)

    ni_zn_charge_cell_volt = FloatType(required=True, min_value=1.2, max_value=2)
    ni_zn_discharge_cell_volt = FloatType(required=True, min_value=0.844, max_value=1.6)
    ni_zn_cell = IntType(required=True, default=0, min_value=0, max_value=10)

    @staticmethod
    def modbus(index, ds1=None, ds2=None, ds3=None, ds4=None, ds5=None):
        if ds1 is not None and ds2 is not None and ds3 is not None and ds4 is not None and ds5 is not None:
            p = Preset()
            p.memory_slot = index
            p.set_from_modbus_data(ds1, ds2, ds3, ds4, ds5)
            return p
        return None

    def verify_can_be_written_or_deleted(self):
        if self.is_fixed:
            # raise ObjectNotFoundException("This preset exists, but is marked as 'fixed' (read only)")
            pass
        return True

    def _test_bit_set(self, offset):
        mask = (1 << offset)
        return (self.op_enable_mask & mask) > 0

    def _set_or_clear_bit(self, offset, bool_value):
        mask = (1 << offset)
        if not bool_value:
            self.op_enable_mask &= ~mask
        else:
            self.op_enable_mask |= mask

    @property
    def is_unused(self):
        return not self.is_used

    @property
    def is_fixed(self):
        return self.use_flag == 0

    @property
    def is_used(self):
        return self.use_flag == 0x55aa

    @property
    def charge_enabled(self):
        return self._test_bit_set(0)

    @charge_enabled.setter
    def charge_enabled(self, new_value):
        self._set_or_clear_bit(0, new_value)

    @property
    def storage_enabled(self):
        return self._test_bit_set(2)

    @storage_enabled.setter
    def storage_enabled(self, new_value):
        self._set_or_clear_bit(2, new_value)

    @property
    def discharge_enabled(self):
        return self._test_bit_set(3)

    @discharge_enabled.setter
    def discharge_enabled(self, new_value):
        self._set_or_clear_bit(3, new_value)

    @property
    def cycle_enabled(self):
        return self._test_bit_set(4)

    @cycle_enabled.setter
    def cycle_enabled(self, new_value):
        self._set_or_clear_bit(4, new_value)

    @property
    def balance_enabled(self):
        return self._test_bit_set(5)

    @balance_enabled.setter
    def balance_enabled(self, new_value):
        self._set_or_clear_bit(5, new_value)

    @property
    def extra_discharge_enable(self):
        return self.li_mode_d & 0x01 == 0x01

    @extra_discharge_enable.setter
    def extra_discharge_enable(self, value):
        if value:
            self.li_mode_d |= 0x01
        else:
            self.li_mode_d &= ~0x01

    @property
    def discharge_balance_enable(self):
        return self.li_mode_d & 0x02 == 0x02

    @discharge_balance_enable.setter
    def discharge_balance_enable(self, value):
        if value:
            self.li_mode_d |= 0x02
        else:
            self.li_mode_d &= ~0x02

    @serializable
    def type_str(self):
        if self.type == 0:
            return "LiPo"
        elif self.type == 1:
            return "LiLo"
        elif self.type == 2:
            return "LiFe"
        elif self.type == 3:
            return "NiMH"
        elif self.type == 4:
            return "Nicd"
        elif self.type == 5:
            return "Pb"

    def to_modbus_data(self):
        new_name = self.name
        if type(new_name) is unicode:
            new_name = new_name.encode('utf8')
        v1 = (self.use_flag,
              new_name,
              self.capacity,
              self.auto_save,
              self.li_balance_end_mode,
              chr(0xff), chr(0xff), chr(0xff), chr(0xff), chr(0xff), chr(0xff), chr(0xff),  # 7 reserved bytes
              self.op_enable_mask,
              self.channel_mode)

        v2 = (self.save_to_sd,
              int(self.log_interval_sec * 10),
              self.run_counter,
              self.type,
              self.li_cell,
              self.ni_cell,
              self.pb_cell,
              self.li_mode_c,
              self.li_mode_d,
              self.ni_mode_c,
              self.ni_mode_d,
              self.pb_mode_c,
              self.pb_mode_d,
              self.bal_speed,
              self.bal_start_mode,
              int(self.bal_start_voltage * 1000),
              self.bal_diff,
              self.bal_over_point,
              self.bal_set_point)

        v3 = (self.bal_delay,
              self.keep_charge_enable,

              int(self.lipo_charge_cell_voltage * 1000),
              int(self.lilo_charge_cell_voltage * 1000),
              int(self.life_charge_cell_voltage * 1000),

              int(self.lipo_storage_cell_voltage * 1000),
              int(self.lilo_storage_cell_voltage * 1000),
              int(self.life_storage_cell_voltage * 1000),

              int(self.lipo_discharge_cell_voltage * 1000),
              int(self.lilo_discharge_cell_voltage * 1000),
              int(self.life_discharge_cell_voltage * 1000),

              int(self.charge_current * 100),
              int(self.discharge_current * 100),
              self.end_charge,
              self.end_discharge,
              self.regen_discharge_mode)

        v4 = (self.ni_peak,
              self.ni_peak_delay,
              self.ni_trickle_enable,
              int(self.ni_trickle_current * 100),
              self.ni_trickle_time,
              self.ni_zero_enable,
              int(self.ni_discharge_voltage * 1000),
              int(self.pb_charge_voltage * 1000),
              int(self.pb_discharge_voltage * 1000),
              self.pb_cell_float_enable,
              int(self.pb_cell_float_voltage * 1000),
              int(self.restore_voltage * 1000),
              self.restore_time,
              int(self.restore_current * 100),
              self.cycle_count,
              self.cycle_delay)

        v5 = (self.cycle_mode,
              self.safety_time_c,
              self.safety_cap_c,
              int(self.safety_temp_c * 10),
              self.safety_time_d,
              self.safety_cap_d,
              int(self.safety_temp_d * 10),
              self.reg_ch_mode,
              int(self.reg_ch_volt * 1000),
              int(self.reg_ch_current * 100),
              self.fast_store,
              self.store_compensation * 100,
              int(self.ni_zn_charge_cell_volt * 1000),
              int(self.ni_zn_discharge_cell_volt * 1000),
              self.ni_zn_cell)

        return v1, v2, v3, v4, v5

    def set_from_modbus_data(self, ds1, ds2, ds3, ds4, ds5):
        (self.use_flag,
         self.name,
         self.capacity,
         self.auto_save,
         self.li_balance_end_mode,
         a, b, c, d, e, f, g,  # 7 reserved bytes
         self.op_enable_mask,
         self.channel_mode) = ds1.data

        (self.save_to_sd,
         self.log_interval_sec,
         self.run_counter,
         self.type,
         self.li_cell,
         self.ni_cell,
         self.pb_cell,
         self.li_mode_c,
         self.li_mode_d,
         self.ni_mode_c,
         self.ni_mode_d,
         self.pb_mode_c,
         self.pb_mode_d,
         self.bal_speed,
         self.bal_start_mode,
         self.bal_start_voltage,
         self.bal_diff,
         self.bal_over_point,
         self.bal_set_point) = ds2.data

        (self.bal_delay,
         self.keep_charge_enable,
         self.lipo_charge_cell_voltage,
         self.lilo_charge_cell_voltage,
         self.life_charge_cell_voltage,
         self.lipo_storage_cell_voltage,
         self.lilo_storage_cell_voltage,
         self.life_storage_cell_voltage,
         self.lipo_discharge_cell_voltage,
         self.lilo_discharge_cell_voltage,
         self.life_discharge_cell_voltage,
         self.charge_current,
         self.discharge_current,
         self.end_charge,
         self.end_discharge,
         self.regen_discharge_mode) = ds3.data

        (self.ni_peak,
         self.ni_peak_delay,
         self.ni_trickle_enable,
         self.ni_trickle_current,
         self.ni_trickle_time,
         self.ni_zero_enable,
         self.ni_discharge_voltage,
         self.pb_charge_voltage,
         self.pb_discharge_voltage,
         self.pb_cell_float_enable,
         self.pb_cell_float_voltage,
         self.restore_voltage,
         self.restore_time,
         self.restore_current,
         self.cycle_count,
         self.cycle_delay) = ds4.data

        (self.cycle_mode,
         self.safety_time_c,
         self.safety_cap_c,
         self.safety_temp_c,
         self.safety_time_d,
         self.safety_cap_d,
         self.safety_temp_d,
         self.reg_ch_mode,
         self.reg_ch_volt,
         self.reg_ch_current,
         self.fast_store,
         self.store_compensation,
         self.ni_zn_charge_cell_volt,
         self.ni_zn_discharge_cell_volt,
         self.ni_zn_cell) = ds5.data

        self.life_charge_cell_voltage /= 1000.0
        self.life_discharge_cell_voltage /= 1000.0
        self.life_storage_cell_voltage /= 1000.0

        self.lilo_charge_cell_voltage /= 1000.0
        self.lilo_discharge_cell_voltage /= 1000.0
        self.lilo_storage_cell_voltage /= 1000.0

        self.lipo_charge_cell_voltage /= 1000.0
        self.lipo_discharge_cell_voltage /= 1000.0
        self.lipo_storage_cell_voltage /= 1000.0

        self.ni_discharge_voltage /= 1000.0

        self.bal_start_voltage /= 1000.0
        self.charge_current /= 100.0
        self.discharge_current /= 100.0

        self.ni_trickle_current /= 100.0
        self.ni_zn_charge_cell_volt /= 1000.0
        self.ni_zn_discharge_cell_volt /= 1000.0

        self.pb_cell_float_voltage /= 1000.0
        self.pb_charge_voltage /= 1000.0
        self.pb_discharge_voltage /= 1000.0

        self.reg_ch_current /= 100
        self.reg_ch_volt /= 1000.0
        self.restore_current /= 100.0
        self.restore_voltage /= 1000.0

        self.log_interval_sec /= 10.0
        self.safety_temp_d /= 10.0
        self.safety_temp_c /= 10.0

        self.store_compensation /= 100.0

        self.name = self.name.split('\0')[0]





import logging
from flask import Flask, Response, request
from flask_cors import CORS
from flask_restful import Api

from rest_interface import \
    DialogCloseResource, \
    StatusResource, \
    SystemStorageResource, \
    ChannelResource, \
    ControlRegisterResource, \
    PresetListResource, \
    PresetResource, ChargeResource, DischargeResource, \
    BalanceResource, MeasureIRResource, StopResource, \
    PresetOrderResource, AddNewPresetResource, StoreResource

application = Flask(__name__, instance_path='/etc')
cors_app = CORS(application)

format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
handler = logging.StreamHandler()
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
application.logger.addHandler(handler)
application.logger.setLevel(logging.DEBUG)

from zmq_marshall import worker_loc
application.logger.info("The charger LIVES! ELECTRIC_WORKER located at: {0}".format(worker_loc))

api = Api(application)
api.add_resource(StatusResource, "/status")
api.add_resource(SystemStorageResource, "/system")
api.add_resource(ControlRegisterResource, "/control")
api.add_resource(ChargeResource, "/charge/<channel_id>/<preset_memory_slot>")
api.add_resource(DischargeResource, "/discharge/<channel_id>/<preset_memory_slot>")
api.add_resource(StoreResource, "/store/<channel_id>/<preset_memory_slot>")
api.add_resource(BalanceResource, "/balance/<channel_id>/<preset_memory_slot>")
api.add_resource(MeasureIRResource, "/measureir/<channel_id>")
api.add_resource(StopResource, "/stop/<channel_id>")
api.add_resource(ChannelResource, "/channel/<channel_id>")
api.add_resource(PresetResource, "/preset/<preset_memory_slot>")
api.add_resource(PresetListResource, "/preset")
api.add_resource(AddNewPresetResource, "/addpreset")
api.add_resource(PresetOrderResource, "/presetorder")
api.add_resource(DialogCloseResource, "/closedialog/<channel_id>")

# application.config['TRAP_HTTP_EXCEPTIONS']=True
#
# @application.errorhandler(Exception)
# def handle_error(e):
#     try:
#         if e.code < 400:
#             return Response.force_type(e, request.environ)
#         elif e.code == 404:
#             return "The page you're looking for was not found", 404
#     except:
#         return make_error_page("Error", "Something went wrong"), 500
#

# @application.errorhandler(ObjectNotFoundException)
# def handle_object_not_found(error):
#     response = jsonify(str(error))
#     response.status_code = error.status_code
#     return response
#
#
# @application.errorhandler(BadRequestException)
# def handle_bad_request(error):
#     response = jsonify(str(error))
#     response.status_code = error.status_code
#     return response




import logging

import modbus_tk.defines as cst

from electric.models import DeviceInfo, ChannelStatus, Control, PresetIndex, Preset, ReadDataSegment, Order, Operation
from electric.models import SystemStorage, WriteDataSegment, OperationResponse, ObjectNotFoundException, \
    BadRequestException, ChemistryType

from modbus_usb import iChargerMaster

CHANNEL_INPUT_HEADER_OFFSET = 0
CHANNEL_INPUT_FOOTER_OFFSET = 51
CHANNEL_INPUT_CELL_IR_FORMAT = 35
CHANNEL_INPUT_CELL_BALANCE_OFFSET = 27
CHANNEL_INPUT_CELL_VOLT_OFFSET = 11

# see the helper/main.cpp module I created that tells me these
# offset values more reliably than Mr Blind Man.
SYSTEM_STORAGE_OFFSET_FANS_OFF_DELAY = 5
SYSTEM_STORAGE_OFFSET_CALIBRATION = 22
SYSTEM_STORAGE_OFFSET_CHARGER_POWER = 34

IR_MEASUREMENT_PRESET_NAME = "IR Measurement"

logger = logging.getLogger('electric.worker.{0}'.format(__name__))

VALUE_ORDER_UNLOCK = 0x55aa
VALUE_ORDER_LOCK = 0x1234 # any other value(s)


class ChargerCommsManager(object):
    """
    The comms manager is responsible for data translation between the MODBUS types and the world outside.  It uses an
    instance of the modbus capable read/write routines to fetch and modify charger parameters.  It co-ordinates
    multiple modbus segment reads/writes where required.

    Validation is not performed here - the data going in/out is assumed to be correct already.
    """
    locking = False

    def __init__(self, master=None):
        if master is None:
            master = iChargerMaster(None)

        self.charger = master

    def reset(self):
        self.charger.reset()

    def turn_off_logging(self):
        values_list = (VALUE_ORDER_UNLOCK, Order.TransLogOff)
        return self.charger.modbus_write_registers(0x8000, values_list)

    def get_device_info(self):
        """
        Returns the following information from the iCharger, known as the 'device only reads message'
        :return: a DeviceInfo instance
        """
        vars = ReadDataSegment(self.charger, "vars", "h12sHHHHHH", base=0x0000)
        return DeviceInfo(vars.data)

    def get_channel_status(self, channel, device_id=None):
        """"
        Returns the following information from the iCharger, known as the 'channel input read only' message:
        :return: ChannelStatus instance
        """
        addr = 0x100 if channel == 0 else 0x200

        # timestamp -> ext temp
        header_fmt = "LlhHHlhh"
        header_data = self.charger.modbus_read_registers(addr, header_fmt)

        # cell 0-15 voltage
        cell_volt_fmt = "16H"
        cell_volt_addr = addr + CHANNEL_INPUT_CELL_VOLT_OFFSET
        cell_volt = self.charger.modbus_read_registers(cell_volt_addr, cell_volt_fmt)

        # cell 0-15 balance
        cell_balance_fmt = "16B"
        cell_balance_addr = addr + CHANNEL_INPUT_CELL_BALANCE_OFFSET
        cell_balance = self.charger.modbus_read_registers(cell_balance_addr, cell_balance_fmt)

        # cell 0-15 IR
        cell_ir_fmt = "16H"
        cell_ir_addr = addr + CHANNEL_INPUT_CELL_IR_FORMAT
        cell_ir = self.charger.modbus_read_registers(cell_ir_addr, cell_ir_fmt)

        # total IR -> dialog box ID
        footer_fmt = "7H"
        footer_addr = addr + CHANNEL_INPUT_FOOTER_OFFSET
        footer = self.charger.modbus_read_registers(footer_addr, footer_fmt)

        return ChannelStatus.modbus(device_id, channel, header_data, cell_volt, cell_balance, cell_ir, footer)

    def get_control_register(self):
        "Returns the current run state of a particular channel"
        return Control(self.charger.modbus_read_registers(0x8000, "7H", function_code=cst.READ_HOLDING_REGISTERS))

    def _beep_summary_dict(self, enabled, volume, type):
        return {
            "enabled": enabled,
            "volume": volume,
            "type": type
        }

    def set_beep_properties(self, beep_index=0, enabled=True, volume=5):
        # for now we only access beep type values
        base = 0x8400

        results = ReadDataSegment(self.charger, "temp", "8H", base=0x8400 + 13)
        value_enabled = list(results.data[:4])
        value_volume = list(results.data[4:])

        value_enabled[beep_index] = int(enabled)
        value_volume[beep_index] = volume

        return self.charger.modbus_write_registers(base + 13, value_enabled + value_volume)

    def set_active_channel(self, channel):
        base = 0x8000 + 2
        if channel not in (0, 1):
            return None
        return self.charger.modbus_write_registers(base, (channel,))

    def get_system_storage(self):
        """Returns the system storage area of the iCharger"""
        # temp-unit -> beep-vol
        ds1 = ReadDataSegment(self.charger, "vars1", "21H", base=0x8400)
        # dump3 -> reg current limit
        ds2 = ReadDataSegment(self.charger, "vars2", "13H", prev_format=ds1)
        # charge/discharge power -> modbus_serial_parity
        ds3 = ReadDataSegment(self.charger, "vars3", "17H", prev_format=ds2)

        return SystemStorage.modbus(ds1, ds2, ds3)

    def save_system_storage(self, system_storage_object):
        (s1, s2, s3, s4, s5, s6) = system_storage_object.to_modbus_data()

        # Write the system data to RAM
        ws1 = self.charger.modbus_write_registers(0x8400, s1)
        ws2 = self.charger.modbus_write_registers(0x8400 + 5, s2)
        ws3 = self.charger.modbus_write_registers(0x8400 + 9, s3)
        ws4 = self.charger.modbus_write_registers(0x8400 + 22, s4)
        ws5 = self.charger.modbus_write_registers(0x8400 + 24, s5)
        ws6 = self.charger.modbus_write_registers(0x8400 + 34, s6)

        # Now write the RAM to flash
        write_sys_to_flash = (VALUE_ORDER_UNLOCK, Order.WriteSys)
        self.charger.modbus_write_registers(0x8000 + 3, write_sys_to_flash)

        return True

    def select_memory_program(self, memory_slot, channel_number=0):
        (word_count,) = self.charger.modbus_write_registers(0x8000 + 1, (memory_slot, channel_number, VALUE_ORDER_UNLOCK))
        return word_count

    def save_full_preset_list(self, preset_list):
        (v1, v2) = preset_list.to_modbus_data()

        # Write the thing to RAM. Moo.
        part1 = WriteDataSegment(self.charger, "part1", v1, "H32B", base=0x8800)
        part2 = WriteDataSegment(self.charger, "part2", v2, "32B", prev_format=part1)

        # Write to flash.
        write_head_to_flash = (VALUE_ORDER_UNLOCK, Order.WriteMemHead)
        self.charger.modbus_write_registers(0x8000 + 3, write_head_to_flash)

        return True

    def find_preset_with_name(self, named):
        for preset in self.get_all_presets():
            preset_name = preset.name
            lowercase_name = str.lower(preset_name)
            lower_case_wanted = str.lower(named)
            if lowercase_name == lower_case_wanted:
                return preset
        return None

    def get_all_presets(self):
        preset_list = self.get_full_preset_list()
        # TODO: Error handling

        all_presets = []
        for index in preset_list.range_of_presets():
            # Preset.index is the memory slot it's in, not the position within the index
            memory_slot_number = preset_list.indexes[index]
            preset = self.get_preset(memory_slot_number)
            if preset.is_used or preset.is_fixed:
                all_presets.append(preset)

        return all_presets

    def get_full_preset_list(self):
        (count,) = self.charger.modbus_read_registers(0x8800, "H", function_code=cst.READ_HOLDING_REGISTERS)

        # There are apparently 64 indexes. Apparently.
        read_a_bit_format = "32B"
        data_1 = self.charger.modbus_read_registers(0x8800 + 1, read_a_bit_format, function_code=cst.READ_HOLDING_REGISTERS)
        data_2 = self.charger.modbus_read_registers(0x8800 + 16, read_a_bit_format, function_code=cst.READ_HOLDING_REGISTERS)
        list_of_all_indexes = list(data_1)
        list_of_all_indexes.extend(list(data_2))

        return PresetIndex.modbus(count, list_of_all_indexes)

    def get_preset(self, memory_slot_number):
        # First, load the indicies and see if in fact this is mapped. If not, don't bother even trying
        preset_index = self.get_full_preset_list()
        if preset_index.index_of_preset_with_memory_slot_number(memory_slot_number) is None:
            message = "there is no preset mapped to slot number {0}".format(memory_slot_number)
            raise ObjectNotFoundException(message)

        result = self.select_memory_program(memory_slot_number)

        # use-flag -> channel mode
        vars1 = ReadDataSegment(self.charger, "vars1", "H38sLBB7cHB", base=0x8c00)
        # save to sd -> bal-set-point
        vars2 = ReadDataSegment(self.charger, "vars2", "BHH12BHBBB", prev_format=vars1)
        # bal-delay, keep-charge-enable -> reg discharge mode
        vars3 = ReadDataSegment(self.charger, "vars3", "BB14H", prev_format=vars2)
        # ni-peak -> cycle-delay
        vars4 = ReadDataSegment(self.charger, "vars4", "16H", prev_format=vars3)
        # cycle-mode -> ni-zn-cell
        vars5 = ReadDataSegment(self.charger, "vars5", "B6HB2HB3HB", prev_format=vars4)

        preset = Preset.modbus(memory_slot_number, vars1, vars2, vars3, vars4, vars5)

        # if preset.is_unused:
        #     message = "Preset in slot {0} appears to exist, is marked as unused.".format(memory_slot_number)
        #     raise ObjectNotFoundException(message)

        return preset

    def delete_preset_at_index(self, preset_memory_slot_number):
        # Find this thing, within the index
        preset_index = self.get_full_preset_list()

        index_number = preset_index.index_of_preset_with_memory_slot_number(preset_memory_slot_number)
        if index_number is None:
            message = "Cannot find preset with memory slot {0}".format(preset_memory_slot_number)
            raise ObjectNotFoundException(message)
        logger.info("Remove item at memory slot {0}, index {1}".format(preset_memory_slot_number, index_number))

        # If the preset is marked as "fixed", then the iCharger UI doesn't allow it to be
        # moved or deleted. Reject the request.
        preset = self.get_preset(preset_memory_slot_number)
        preset.verify_can_be_written_or_deleted()

        # If it is the last object, we can adjust the index map only, and ignore (I hope!) the preset itself.
        preset_index.delete_item_at_index(index_number)

        # Change the preset index so that the last item is "unused"
        index_save_result = self.save_full_preset_list(preset_index)

        # Set this preset (slot) used flag to "EMPTY (useflag = 0xffff") in RAM
        logger.info("Setting slot {0} unused flag".format(preset_memory_slot_number))
        self.select_memory_program(preset_memory_slot_number)
        self.charger.modbus_write_registers(0x8c00, (0xffff,))

        # Now write back to flash
        write_to_flash = (VALUE_ORDER_UNLOCK, Order.WriteMem)
        self.charger.modbus_write_registers(0x8000 + 3, write_to_flash)

        return True

    '''
    This ALWAYS saves a NEW preset. The presets memory_slot is ignored, and it's
    inserted at the end of the preset index list.
    '''
    def add_new_preset(self, preset):
        # Find the next free memory slot, assign that to the preset, and save both indexes + preset
        preset_index = self.get_full_preset_list()

        # This assigns the preset the next available memory slot, and also writes that into
        # the index.
        if not preset_index.add_to_index(preset):
            raise BadRequestException("Presets full")

        # By default, presets should NOT have auto_save enabled
        preset.auto_save = False

        # Right. Now we can save it.
        logger.info("Preset Index before add: {0}".format(preset_index.to_native()))
        logger.info("Adding new preset at slot {0}".format(preset.memory_slot))
        logger.info("The preset being sent in is: {0}".format(preset))

        self.save_preset_to_memory_slot(preset, preset.memory_slot, verify_write=False)

        # And save the new preset list
        self.save_full_preset_list(preset_index)
        logger.info("Preset Index after add: {0}".format(self.get_full_preset_list().to_native()))

        returned_preset = self.get_preset(preset.memory_slot)
        logger.info("The preset being read out is: {0}".format(returned_preset))
        return returned_preset

    '''
    This saves an existing preset to memory.
    It does NOT allocate new presets, or insert them into a preset index list
    '''
    def save_preset_to_memory_slot(self, preset, memory_slot, write_to_flash=True, verify_write=True):
        # We don't want to verify if we're adding. In that case we KNOW we want to add it here.
        # and we want to ignore any older data that may be in that slot.
        if verify_write:
            # Get the preset and verify we can write
            # This has a side effect of self.select_memory_program(memory_slot)
            existing_preset = self.get_preset(memory_slot)
            existing_preset.verify_can_be_written_or_deleted()
        else:
            self.select_memory_program(memory_slot)

        # ask the preset for its data segments
        (v1, v2, v3, v4, v5) = preset.to_modbus_data()

        # Write the preset into the RAM area
        s1 = WriteDataSegment(self.charger, "seg1", v1, "H38sLBB7cHB", base=0x8c00)
        s2 = WriteDataSegment(self.charger, "seg2", v2, "BHH12BHBBB", prev_format=s1)
        s3 = WriteDataSegment(self.charger, "seg3", v3, "BB14H", prev_format=s2)
        s4 = WriteDataSegment(self.charger, "seg4", v4, "16H", prev_format=s3)
        s5 = WriteDataSegment(self.charger, "seg5", v5, "B6HB2HB3HB", prev_format=s4)

        # Now write back to flash
        if write_to_flash:
            write_to_flash = (VALUE_ORDER_UNLOCK, Order.WriteMem)
            self.charger.modbus_write_registers(0x8000 + 3, write_to_flash)

        return True

    def close_messagebox(self, channel_id):
        # no-op
        pass

    def stop_operation(self, channel_number):
        channel_number = min(1, max(0, channel_number))
        values_list = (channel_number, VALUE_ORDER_UNLOCK, Order.Stop)

        modbus_response = self.charger.modbus_write_registers(0x8000 + 2, values_list)

        # status = self.get_device_info().get_status(channel_number)
        # logger.info("Device status: {0}".format(status.to_native()))

        # If showing a dialog, or have error, try to clear the dialog
        # This also works to close the presets listing, if that is open
        # if status.dlg_box_status or status.err:
        #     self.close_messagebox(channel_number)

        return OperationResponse(modbus_response)

    def run_operation(self, operation, channel_number, preset_memory_slot_index=0):
        if operation == Operation.MeasureIR:
            return self.measure_ir(channel_number)

        channel_number = min(1, max(0, channel_number))

        # # Load the preset from this slot, and check it is valid
        # # Loading will itself perform a check for 'used', and will throw an exception if it is not available.
        # # It'll also be loaded into RAM.
        # self.get_preset(preset_memory_slot_index)

        values_list = (operation, preset_memory_slot_index, channel_number, VALUE_ORDER_UNLOCK, Order.Run)

        # logger.info("Sending run_op to channel {2}: op {0}, preset: {1}".format(operation, preset_memory_slot_index, channel_number))
        self.charger.modbus_write_registers(0x8000, values_list)

        return self.get_device_info().get_status(channel_number)


    def measure_ir(self, channel_number):
        channel_number = min(1, max(0, channel_number))

        # Kick off a discharge, 2A.
        # First, pick a suitable LIPO preset.
        preset_index = self.get_full_preset_list()
        preset_for_ir = self.find_preset_with_name(IR_MEASUREMENT_PRESET_NAME)

        if not preset_for_ir:
            for memory_slot in preset_index.range_of_presets():
                preset = self.get_preset(memory_slot)
                if preset.type == ChemistryType.LiPo:
                    preset_for_ir = preset
                    break

            if not preset_for_ir:
                raise ObjectNotFoundException("Could not find any Lipo preset to use for IR measurement")

            # Copy this into its own slot.
            logger.info("Creating a new preset specifically for IR usage")
            preset_for_ir.discharge_current = 2
            preset_for_ir.safety_time_d = 1
            preset_for_ir.name = IR_MEASUREMENT_PRESET_NAME
            preset_for_ir = self.add_new_preset(preset_for_ir)

        logger.info("Using preset at slot {0} for IR measure".format(preset_for_ir.memory_slot))

        # Perform a discharge until we see 2A. Maximum time will be 25s.
        self.run_operation(Operation.Discharge, channel_number, preset_for_ir.memory_slot)

        # Problem. This should be Async...
        # Restore original preset in RAM (possibly not actually necessary)
        # self.save_preset_to_memory_slot(original_preset, original_preset.index, write_to_flash=False, verify_write=False)

        return self.get_device_info().get_status(channel_number)



from electric.zmq_marshall import ZMQCommsManager
import zmq.utils.win32, logging, sys, time

from electric.worker.comms_layer import Operation

logger = logging.getLogger('electric.test')
comms = ZMQCommsManager()

def stop_application():
    sys.exit(0)

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

with zmq.utils.win32.allow_interrupt(stop_application):
    print comms.get_device_info()
    print comms.get_control_register()
    print comms.get_system_storage()
    print comms.get_full_preset_list()
    print comms.set_active_channel(0)
    print comms.set_active_channel(1)
    print comms.get_channel_status(0, None)
    print comms.get_channel_status(1, None)
    print comms.select_memory_program(0, 0)
    print comms.select_memory_program(0, 1)
    print comms.get_preset(0)
    print comms.get_preset(1)
    print comms.run_operation(Operation.Charge, 0, 0)
    time.sleep(5)
    print comms.stop_operation(0)
    print comms.stop_operation(0)
    print comms.measure_ir(0)
    time.sleep(5)
    print comms.stop_operation(0)
    print comms.stop_operation(0)
